"""Module that holds helper routines for the graph module.  This module
is not part of the wnnet API.  This documentation is shown simply for
completeness."""

import operator
from math import floor, log10
import networkx as nx


def get_keywords(kw_list, **kwargs):
    """Helper function to return keyword arguments for graph module."""

    for key in kwargs:
        assert key in kw_list, f"Keyword {key} not allowed in this routine."

    def_kwargs = {}

    def_kwargs["induced_nuc_xpath"] = ""
    def_kwargs["induced_reac_xpath"] = ""
    def_kwargs["flow_type"] = "net"
    def_kwargs["direction"] = "both"
    def_kwargs["reaction_color_tuples"] = None
    def_kwargs["user_funcs"] = None
    def_kwargs["zone_user_funcs"] = None
    def_kwargs["threshold"] = 0.01
    def_kwargs["scale"] = 10
    def_kwargs["state_scaling"] = 0.35
    def_kwargs["allow_isolated_species"] = False
    def_kwargs["title_func"] = None
    def_kwargs["zone_title_func"] = None
    def_kwargs["node_label_func"] = None
    def_kwargs["zone_node_label_func"] = None
    def_kwargs["scale_edge_weight_func"] = None
    def_kwargs["graph_attributes"] = {"outputorder": "edgesfirst"}
    def_kwargs["node_attributes"] = {
        "shape": "box",
        "fontsize": 16,
        "style": "filled",
        "fillcolor": "white",
    }
    def_kwargs["edge_attributes"] = {"arrowsize": 0.2}
    def_kwargs["solar_species"] = get_solar_species_list()
    def_kwargs["solar_node_attributes"] = {
        "fillcolor": "yellow",
        "style": "filled",
    }
    def_kwargs["special_node_attributes"] = None
    def_kwargs["special_edge_attributes"] = None

    result = {}

    for k_w in kw_list:
        if k_w in kwargs:
            if isinstance(def_kwargs[k_w], dict):
                result[k_w] = {**def_kwargs[k_w], **kwargs[k_w]}
            else:
                result[k_w] = kwargs[k_w]
        else:
            result[k_w] = def_kwargs[k_w]

    return result


def get_solar_species_list():
    """Helper function to return list of solar species."""

    return [
        "h1",
        "h2",
        "he3",
        "he4",
        "li6",
        "li7",
        "be9",
        "b10",
        "b11",
        "c12",
        "c13",
        "n14",
        "n15",
        "o16",
        "o17",
        "o18",
        "f19",
        "ne20",
        "ne21",
        "ne22",
        "na23",
        "mg24",
        "mg25",
        "mg26",
        "al27",
        "si28",
        "si29",
        "si30",
        "p31",
        "s32",
        "s33",
        "s34",
        "s36",
        "cl35",
        "cl37",
        "ar36",
        "ar38",
        "ar40",
        "k39",
        "k40",
        "k41",
        "ca40",
        "ca42",
        "ca43",
        "ca44",
        "ca46",
        "ca48",
        "sc45",
        "ti46",
        "ti47",
        "ti48",
        "ti49",
        "ti50",
        "v50",
        "v51",
        "cr50",
        "cr52",
        "cr53",
        "cr54",
        "mn55",
        "fe54",
        "fe56",
        "fe57",
        "fe58",
        "co59",
        "ni58",
        "ni60",
        "ni61",
        "ni62",
        "ni64",
        "cu63",
        "cu65",
        "zn64",
        "zn66",
        "zn67",
        "zn68",
        "zn70",
        "ga69",
        "ga71",
        "ge70",
        "ge72",
        "ge73",
        "ge74",
        "ge76",
        "as75",
        "se74",
        "se76",
        "se77",
        "se78",
        "se80",
        "se82",
        "br79",
        "br81",
        "kr78",
        "kr80",
        "kr82",
        "kr83",
        "kr84",
        "kr86",
        "rb85",
        "rb87",
        "sr84",
        "sr86",
        "sr87",
        "sr88",
        "y89",
        "zr90",
        "zr91",
        "zr92",
        "zr94",
        "zr96",
        "nb93",
        "mo92",
        "mo94",
        "mo95",
        "mo96",
        "mo97",
        "mo98",
        "mo100",
        "ru96",
        "ru98",
        "ru99",
        "ru100",
        "ru101",
        "ru102",
        "ru104",
        "rh103",
        "pd102",
        "pd104",
        "pd105",
        "pd106",
        "pd108",
        "pd110",
        "ag107",
        "ag109",
        "cd106",
        "cd108",
        "cd110",
        "cd111",
        "cd112",
        "cd113",
        "cd114",
        "cd116",
        "in113",
        "in115",
        "sn112",
        "sn114",
        "sn115",
        "sn116",
        "sn117",
        "sn118",
        "sn119",
        "sn120",
        "sn122",
        "sn124",
        "sb121",
        "sb123",
        "te120",
        "te122",
        "te123",
        "te124",
        "te125",
        "te126",
        "te128",
        "te130",
        "i127",
        "xe124",
        "xe126",
        "xe128",
        "xe129",
        "xe130",
        "xe131",
        "xe132",
        "xe134",
        "xe136",
        "cs133",
        "ba130",
        "ba132",
        "ba134",
        "ba135",
        "ba136",
        "ba137",
        "ba138",
        "la138",
        "la139",
        "ce136",
        "ce138",
        "ce140",
        "ce142",
        "pr141",
        "nd142",
        "nd143",
        "nd144",
        "nd145",
        "nd146",
        "nd148",
        "nd150",
        "sm144",
        "sm147",
        "sm148",
        "sm149",
        "sm150",
        "sm152",
        "sm154",
        "eu151",
        "eu153",
        "gd152",
        "gd154",
        "gd155",
        "gd156",
        "gd157",
        "gd158",
        "gd160",
        "tb159",
        "dy156",
        "dy158",
        "dy160",
        "dy161",
        "dy162",
        "dy163",
        "dy164",
        "ho165",
        "er162",
        "er164",
        "er166",
        "er167",
        "er168",
        "er170",
        "tm169",
        "yb168",
        "yb170",
        "yb171",
        "yb172",
        "yb173",
        "yb174",
        "yb176",
        "lu175",
        "lu176",
        "hf174",
        "hf176",
        "hf177",
        "hf178",
        "hf179",
        "hf180",
        "ta180",
        "ta181",
        "w180",
        "w182",
        "w183",
        "w184",
        "w186",
        "re185",
        "re187",
        "os184",
        "os186",
        "os187",
        "os188",
        "os189",
        "os190",
        "os192",
        "ir191",
        "ir193",
        "pt190",
        "pt192",
        "pt194",
        "pt195",
        "pt196",
        "pt198",
        "au197",
        "hg196",
        "hg198",
        "hg199",
        "hg200",
        "hg201",
        "hg202",
        "hg204",
        "tl203",
        "tl205",
        "pb204",
        "pb206",
        "pb207",
        "pb208",
        "bi209",
        "th232",
        "u235",
        "u238",
    ]


def color_edges(my_graph, net, color_tuples):
    """Helper function to color arcs in graph by reaction type."""
    color = {}
    for reaction in net.get_reactions():
        color[reaction] = "black"

    if color_tuples:
        for color_tup in color_tuples:
            for reaction in net.get_reactions(reac_xpath=color_tup[0]):
                color[reaction] = color_tup[1]

        for edge in my_graph.edges:
            if not "color" in my_graph.edges[edge]:
                my_graph.edges[edge]["color"] = color[edge[2]]


def fexp(_x):
    """Helper function to compute an exponent of a number."""
    # Add 0.01 for rounding for :.2f mantissa formating
    return int(floor(log10(abs(_x)) + 0.01)) if _x != 0.0 else 0


def fman(_x):
    """Helper function to compute the mantissa of a number."""
    return _x / 10.0 ** fexp(_x)


def get_pos(coll, name, state_scaling):
    """Helper function to define the position of a node in a network graph."""
    _z, _a, state = coll.xml.get_z_a_state_from_nuclide_name(name)
    _n = _a - _z
    if state == "g":
        _z -= state_scaling
    elif state == "m":
        _z += state_scaling
    return (float(_n), float(_z))


def get_subset_and_anchors(nuclides):
    """Helper function to define the scope and anchors of a graph."""
    species_subset = []
    my_dict = {}
    z_dict = {}
    for key, value in nuclides.items():
        species_subset.append(key)
        my_dict[(value["z"], value["a"])] = key
        if value["z"] not in z_dict:
            z_dict[value["z"]] = [value["a"]]
        else:
            z_dict[value["z"]].append(value["a"])

    z_array = []

    for key, value in z_dict.items():
        z_array.append(key)
        value.sort()

    z_array.sort()

    anchors = []
    anchors.append(my_dict[(z_array[0], z_dict[z_array[0]][0])])
    anchors.append(my_dict[(z_array[0], z_dict[z_array[0]][-1])])
    anchors.append(my_dict[(z_array[-1], z_dict[z_array[-1]][0])])
    anchors.append(my_dict[(z_array[-1], z_dict[z_array[-1]][-1])])

    return (species_subset, anchors)


def apply_graph_attributes(my_graph, graph_attributes):
    """Helper function to apply attributes to the graph as a whole."""
    if graph_attributes:
        for key, value in graph_attributes.items():
            my_graph.graph[key] = value


def apply_node_attributes(my_graph, node_attributes):
    """Helper function to apply attributes to the nodes of a graph."""
    if node_attributes:
        for key, value in node_attributes.items():
            for node in my_graph.nodes:
                my_graph.nodes[node][key] = value


def apply_edge_attributes(my_graph, edge_attributes):
    """Helper function to apply attributes to the edges of a graph."""
    if edge_attributes:
        for key, value in edge_attributes.items():
            for edge in my_graph.edges:
                my_graph.edges[edge][key] = value


def apply_solar_node_attributes(
    my_graph, solar_species, solar_node_attributes
):
    """Helper function to apply attributes to the nodes representing solar\
       species in a graph."""

    if solar_node_attributes:
        for node in solar_species:
            if node in my_graph.nodes:
                for key, value in solar_node_attributes.items():
                    my_graph.nodes[node][key] = value


def apply_special_node_attributes(my_graph, special_node_attributes):
    """Helper function to apply attributes to special nodes in a graph."""
    if special_node_attributes:
        for node in special_node_attributes:
            for key, value in special_node_attributes[node].items():
                my_graph.nodes[node][key] = value


def apply_special_edge_attributes(my_graph, special_edge_attributes):
    """Helper function to apply attributes to special edges in a graph."""
    if special_edge_attributes:
        for edge in special_edge_attributes:
            if not edge in my_graph.edges:
                my_graph.add_edge(*edge)
            for key, value in special_edge_attributes[edge].items():
                my_graph.edges[edge][key] = value


def create_flow_graph(net, my_flows, subset_nuclides, anchors, **my_args):
    """Helper function to create a flow graph."""

    d_g = nx.MultiDiGraph()

    for nuc in net.get_nuclides():
        d_g.add_node(nuc)

    add_flows_to_graph(net, d_g, my_flows, my_args)

    # Apply attributes

    apply_graph_attributes(d_g, my_args["graph_attributes"])

    apply_node_attributes(d_g, my_args["node_attributes"])

    apply_edge_attributes(d_g, my_args["edge_attributes"])

    apply_solar_node_attributes(
        d_g, my_args["solar_species"], my_args["solar_node_attributes"]
    )

    apply_special_node_attributes(d_g, my_args["special_node_attributes"])

    # Subgraph and widths.

    sub_graph = nx.subgraph(d_g, subset_nuclides)

    f_max = set_widths_and_get_max_weight(d_g, sub_graph, my_args)

    # Apply special edge attributes after scaling.

    apply_special_edge_attributes(d_g, my_args["special_edge_attributes"])

    # Remove isolated nodes if desired

    if not my_args["allow_isolated_species"]:
        remove_isolated_nodes(d_g, my_args)

    # Restore anchors

    for anchor in anchors:
        if anchor not in d_g.nodes:
            d_g.add_node(anchor, style="invis")

    # Get new subset

    sub_graph_2 = nx.subgraph(d_g, subset_nuclides)

    for node in sub_graph_2.nodes:
        sub_graph_2.nodes[node]["pos"] = get_pos(
            net, node, my_args["state_scaling"]
        )
        sub_graph_2.nodes[node]["label"] = my_args["node_label_func"](node)

    # Title

    d_g.graph["label"] = my_args["title_func"](f_max)

    color_edges(sub_graph_2, net, my_args["reaction_color_tuples"])

    return sub_graph_2


def create_integrated_current_graph(
    net, zone, subset_nuclides, anchors, **my_args
):
    """Helper function to create an integrated current graph."""

    # Create graph.

    d_g = nx.MultiDiGraph()

    # Add nodes and arcs.

    for nuc in net.get_nuclides():
        d_g.add_node(nuc)

    add_currents_to_graph(net, zone, d_g)

    # Apply attributes

    apply_graph_attributes(d_g, my_args["graph_attributes"])

    apply_node_attributes(d_g, my_args["node_attributes"])

    apply_edge_attributes(d_g, my_args["edge_attributes"])

    apply_solar_node_attributes(
        d_g, my_args["solar_species"], my_args["solar_node_attributes"]
    )

    apply_special_node_attributes(d_g, my_args["special_node_attributes"])

    # Subgraph and maximum flow within subgraph

    sub_graph = nx.subgraph(d_g, subset_nuclides)

    f_max = set_widths_and_get_max_weight(d_g, sub_graph, my_args)

    # Apply special edge attributes after scaiing.

    apply_special_edge_attributes(d_g, my_args["special_edge_attributes"])

    # Remove isolated nodes if desired

    if not my_args["allow_isolated_species"]:
        remove_isolated_nodes(d_g, my_args)

    # Restore anchors

    for anchor in anchors:
        if anchor not in d_g.nodes:
            d_g.add_node(anchor, style="invis")

    # Get new subset

    sub_graph_2 = nx.subgraph(d_g, subset_nuclides)

    for node in sub_graph_2.nodes:
        sub_graph_2.nodes[node]["pos"] = get_pos(
            net, node, my_args["state_scaling"]
        )
        sub_graph_2.nodes[node]["label"] = my_args["node_label_func"](node)

    d_g.graph["label"] = my_args["title_func"](f_max)

    color_edges(sub_graph_2, net, my_args["reaction_color_tuples"])

    return sub_graph_2


def add_reactions_to_graph(net, d_g, my_args):
    """Helper function to add reaction arcs to graph."""

    reactions = net.get_reactions(reac_xpath=my_args["induced_reac_xpath"])

    for key, value in reactions.items():
        if my_args["direction"] in ("forward", "both"):
            for reactant in value.nuclide_reactants:
                for product in value.nuclide_products:
                    d_g.add_edge(reactant, product, key)

        if not net.is_weak_reaction(key) and (
            my_args["direction"] in ("reverse", "both")
        ):
            for product in value.nuclide_products:
                for reactant in value.nuclide_reactants:
                    d_g.add_edge(product, reactant, key)


def add_flows_to_graph(net, d_g, my_flows, my_args):
    """Helper function to add flow arcs to graph."""

    reactions = net.get_reactions()

    for key, value in my_flows.items():

        if my_args["flow_type"] == "full":

            if value[0] > 0:
                add_flow_edges(
                    d_g,
                    reactions[key].nuclide_reactants,
                    reactions[key].products,
                    value[0],
                    key,
                )

            if value[1] > 0:
                add_flow_edges(
                    d_g,
                    reactions[key].nuclide_products,
                    reactions[key].reactants,
                    value[1],
                    key,
                )

        if my_args["flow_type"] == "net":

            net_flow = value[0] - value[1]

            if net_flow > 0:
                add_flow_edges(
                    d_g,
                    reactions[key].nuclide_reactants,
                    reactions[key].nuclide_products,
                    net_flow,
                    key,
                )
            else:
                add_flow_edges(
                    d_g,
                    reactions[key].nuclide_products,
                    reactions[key].nuclide_reactants,
                    -net_flow,
                    key,
                )


def add_currents_to_graph(net, zone, d_g):
    """Helper function to add current arcs to graph."""

    reactions = net.get_reactions()

    props = zone["properties"]

    f_currents = {}

    for prop in props:
        if isinstance(prop, tuple):
            if prop[0] == "flow current":
                f_currents[prop[1]] = float(props[prop])

    for key, value in f_currents.items():
        if value > 0:
            for reactant in reactions[key].nuclide_reactants:
                for product in reactions[key].nuclide_products:
                    d_g.add_edge(reactant, product, key, weight=value)
        else:
            for product in reactions[key].nuclide_products:
                for reactant in reactions[key].nuclide_reactants:
                    d_g.add_edge(product, reactant, key, weight=-value)


def remove_isolated_nodes(d_g, my_args):
    """Helper function to remove isolated nodes."""

    isolated_nodes = list(nx.isolates(d_g))
    for node in isolated_nodes:
        if node not in my_args["solar_species"] and (
            not my_args["special_node_attributes"]
            or node not in my_args["special_node_attributes"]
        ):
            d_g.remove_node(node)


def set_widths_and_get_max_weight(d_g, sub_graph, my_args):
    """Helper function to set edge widths or remove."""

    my_weights = nx.get_edge_attributes(sub_graph, "weight")

    f_max = 0

    if len(my_weights) > 0:
        f_max = max(my_weights.items(), key=operator.itemgetter(1))[1]

        remove_edges = []
        for edge in d_g.edges:
            if not my_args["scale_edge_weight_func"](
                d_g.get_edge_data(*edge),
                f_max,
                my_args["scale"],
                my_args["threshold"],
            ):
                remove_edges.append(edge)

        d_g.remove_edges_from(remove_edges)

    return f_max


def add_flow_edges(d_g, s_array, t_array, my_weight, key):
    """Helper function to add flow edges."""

    for source in s_array:
        for target in t_array:
            d_g.add_edge(source, target, key, weight=my_weight)



"""A module to create graphs from network data.

.. _Allowed_keywords:

The routines in this module allow for a number of optional keyword
arguments.  Not all keyword arguments are available for each routine,
however.  The documentation indicates which keyword arguments are
appropriate for a given routine.

The possible keyword arguments are:

.. _induced_nuc_xpath:

    ``induced_nuc_xpath`` (:obj:`str`): An XPath expression to select\
        the subset of nuclides in the graph.  The default is all species\
        in the network.

.. _induced_reac_xpath:

    ``induced_reac_xpath`` (:obj:`str`): An XPath expression to select\
        the subset of reactions in the graph.  The default is all reactions\
        in the network.

.. _flow_type:

    ``flow_type`` (:obj:`str`) A string giving the flow type to be\
        presented.  The possible values are `net`, which shows the forward\
        minus the reverse flow (or the opposite if the reverse flow is larger),\
        and `full`, which shows both the foward and reverse flows.  The\
        default is `net`.

.. _direction:

    ``direction`` (:obj:`str`) A string indicting the reaction directions\
        to show.  Allowed values are `forward`, `reverse`, and `both`\
        (the default, which shows both `forward` and `reverse`).

.. _reaction_color_tuples:

    ``reaction_color_tuples`` (:obj:`list`): A list of :obj:`tuple` objects\
        to select arc colors.  There is a tuple for each reaction type.\
        The first member of the tuple is an XPath expression to select the\
        reaction type while the second member is a string giving the color\
        for that reaction type.  The default is that all arcs are black.

.. _user_funcs:

    ``user_funcs`` (:obj:`dict`): A dictionary of user-defined\
        functions associated with a user_rate key.  The prototype for each\
        user rate function should be (*reaction*, *t_9*), where\
        *t_9* is the temperature in billions of Kelvin and *reaction*\
        is a `wnutils <https://wnutils.readthedocs.io>`_ reaction\
        instance.  Other data can be bound to the function.

.. _zone_user_funcs:

     ``zone_user_funcs`` (:obj:`dict`): A dictionary of user-defined\
        functions associated with a user_rate key.\
        The prototype for each\
        user rate function should be (*reaction*, *t_9*, *zone*), where\
        *t_9* is the temperature in billions of Kelvin and *reaction* and\
        *zone* are `wnutils <https://wnutils.readthedocs.io>`_ reaction and\
        zone instances.  Other data can be bound to the function.

.. _threshold:

    ``threshold`` (:obj:`float`):  The minimum flow (relative to the\
        maximum flow) to be shown on the graph.  The default is 0.01.

.. _scale:
        
    ``scale`` (:obj:`float`):  Scaling factor for the maximum weight arc.\
        The default is 10.

.. _state_scaling:
        
    ``state_scaling`` (:obj:`float`):  Scaling factor for isomeric states.\
        The default is 0.35.

.. _allow_isolated_species:
        
    ``allow_isolated_species`` (:obj:`bool`):  Boolean to choose whether\
        to allow isolated species (ones without incoming or outgoing arcs)\
        in the graph.  The default is *False*.

.. _title_func:

    ``title_func``: A\
       `function <https://docs.python.org/3/library/stdtypes.html#functions>`_\
        that applies the title to the graph.  The function must take\
        one :obj:`float` argument giving the maximum flow.\
        Other data can be bound to the function.  The function must return\
        a :obj:`str` giving the title.  The default is\
        :meth:`wnnet.graph.make_t9_rho_flow_string`.

.. _zone_title_func:
 
    ``zone_title_func``: A\
        `function <https://docs.python.org/3/library/stdtypes.html#functions>`_\
        that applies the title to the graph.  The function must take\
        three arguments.  The first is a :obj:`float` giving the maximum\
        flow.  The second is the zone object corresponding to\
        the graph while the third is the zone label.  Other data can \
        be bound to the function.  The function must return a :obj:`str` \
        giving the title.\
        The default is :meth:`wnnet.graph.make_time_t9_rho_flow_string`\
        for flow graphs and\
        :meth:`wnnet.graph.make_time_t9_rho_current_string` for integrated\
        current graphs.

.. _node_label_func:
        
    ``node_label_func``: A\
      `function  <https://docs.python.org/3/library/stdtypes.html#functions>`_ \
       that applies a label to each node in the graph.  The function\
       must take as argument a species name.  Other data can be bound to\
       the function.  The function must return a :obj:`str`\
       giving the label.\
       The default is :meth:`wnnet.graph.make_node_label`.  The *g_names*\
       argument for the default is bound by the calling routine.

.. _zone_node_label_func:

    ``zone_node_label_func``: A `function \
        <https://docs.python.org/3/library/stdtypes.html#functions>`_ \
        that applies a label to each node in the graph.  The function \
        must take as arguments a species name, a zone, and the zone label. \
        Other data can be bound to \
        the function.  The function must return a :obj:`str` \
        giving the label.  The default is \
        :meth:`wnnet.graph.make_zone_node_label`.

.. _scale_edge_weight_func:
        
    ``scale_edge_weight_func``: A `function \
        <https://docs.python.org/3/library/stdtypes.html#functions>`_ \
        that applies scales each edge weight in the graph.  The function \
        must take four arguments: a dictionary of edge data, the \
        maximum edge weight in the scope of the graph, a scale factor \
        by which to scale the weight (input as `scale`_ to this routine), \
        and a threshold for not including the edge in the graph \
        (input as `threshold`_ to this routine). \
        Other data can be bound to \
        the function.  The function must modify the weight in the \
        edge data and return a :obj:`bool` indicating whether to include \
        the edge in the graph (True) or not (False).\
        The default is :meth:`wnnet.graph.scale_edge_weight`.

.. _graph_attributes:
        
    ``graph_attributes`` (:obj:`dict`):  A dictionary of graphviz attributes\
        for the graph.  The default is *{"outputorder": "edgesfirst"}*.

.. _edge_attributes:
        
    ``edge_attributes`` (:obj:`dict`):  A dictionary of grapvhiz attributes\
        for the edges in the graph.  The default is *{"arrowsize": 0.2}*.

.. _node_attributes:
        
    ``node_attributes`` (:obj:`dict`):  A dictionary of graphviz\
        attributes for the nodes in the graph.  The default is\
        *{ "shape": "box", "fontsize": 16, "style": "filled",\
         "fillcolor": "white" }*.

.. _solar_species:

    ``solar_species`` (:obj:`list`):  A list of species to be\
        considered as the naturally occurring species.  The default is the list\
        returned from :meth:`wnnet.graph.get_solar_species`.

.. _solar_node_attributes:

    ``solar_node_attributes`` (:obj:`dict`):  A dictionary of graphviz\
        attributes to be applied to the solar species in the graph.\
        The default is *{"fillcolor": "yellow", "style": "filled"}*.

.. _special_node_attributes:

    ``special_node_attributes`` (:obj:`dict`):  A dictionary of graphviz\
        attributes to be applied to the special nodes in the graph.\
        The dictionary has as keys the names of the special nodes and as values\
        a dictionary of graphviz properties to be applied to the given special\
        node.

.. _special_edge_attributes:

    ``special_edge_attributes`` (:obj:`dict`):  A dictionary of graphviz\
        attributes to be applied to the special edges in the graph.\
        The dictionary has as keys a :obj:`tuple` *(u, v, reaction)*,\
        where *u* is the source of the directed edge (i.e., arc),\
        *v* is the target,\
        and *reaction* is the reaction from which the arc derives.  The\
        *reaction* tuple element serves as a key in the multi-digraph to\
        distinguish parallel arcs.  The values of the dictionary are each\
        a dictionary of graphviz properties to be applied to the given special\
        edge.

"""

import functools
import networkx as nx
import wnnet.flows as wf
import wnnet.graph_helper as gh


def get_solar_species():
    """A method to return the naturally-occurring solar-system species.

    Returns:
        A :obj:`list`.  A list containing :obj:`str` of the names
        of the naturally-occurring solar-system species.

    """

    return gh.get_solar_species_list()


def make_time_t9_rho_flow_string(f_max, zone, zone_label):
    """The default title function for zone flow graphs.

    Args:
        ``f_max`` (:obj:`float`):  The maximum flow in the scope of the graph.

        ``zone``:  A `wnutils <https://wnutils.readthedocs.io>`_ zone object.

        ``zone_label`` (:obj:`str` or :obj:`tuple`):  The label of the zone.

    Returns:
        A :obj:`str`.  The title string.

    """

    assert zone_label

    props = zone["properties"]
    time = float(props["time"])
    t_9 = float(props["t9"])
    rho = float(props["rho"])

    return f"<time (s) = {gh.fman(time):.2f} x 10<sup>{gh.fexp(time):d}</sup>, \
T<sub>9</sub> = {t_9:.2f}, \
rho (g/cc) = {gh.fman(rho):.2f} x 10<sup>{gh.fexp(rho):d}</sup> (g/cc), \
max. flow = {gh.fman(f_max):.2f} x 10<sup>{gh.fexp(f_max):d}</sup> (s<sup>-1</sup>)>"


def make_time_t9_rho_current_string(f_max, zone, zone_label):
    """The default title function for zone integrated current graphs.

    Args:
        ``f_max`` (:obj:`float`):  The maximum integrated current in the scope\
          of the graph.

        ``zone``:  A `wnutils <https://wnutils.readthedocs.io>`_ zone object.

        ``zone_label`` (:obj:`str` or :obj:`tuple`):  The label of the zone.

    Returns:
        A :obj:`str`.  The title string.

    """

    assert zone_label

    props = zone["properties"]
    time = float(props["time"])
    t_9 = float(props["t9"])
    rho = float(props["rho"])

    return f"<time (s) = {gh.fman(time):.2f} x 10<sup>{gh.fexp(time):d}</sup>, \
T<sub>9</sub> = {t_9:.2f}, \
rho (g/cc) = {gh.fman(rho):.2f} x 10<sup>{gh.fexp(rho):d}</sup> (g/cc), \
max. int. current = {gh.fman(f_max):.2f} x 10<sup>{gh.fexp(f_max):d}</sup>>"


def make_t9_rho_flow_string(f_max, t_9, rho):
    """The default title function for flow graphs.

    Args:
        ``f_max`` (:obj:`float`):  The maximum flow in the scope of the graph.

        ``t_9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which
        the flows are computed.  This value is bound by the calling function.

        ``rho`` (:obj:`float`):  The density in g/cc at which the flows are
        computed.  This value is bound by the calling function.

    Returns:
        A :obj:`str`.  The title string.

    """

    return f"<T<sub>9</sub> = {t_9:.2f}, \
rho (g/cc) = {gh.fman(rho):.2f} x 10<sup>{gh.fexp(rho):d}</sup> (g/cc), \
max. flow = {gh.fman(f_max):.2f} x 10<sup>{gh.fexp(f_max):d}</sup> (s<sup>-1</sup>)>"


def make_node_label(name, g_names):
    """The default node label function.

    Args:
        ``name`` (:obj:`str`):  The species name.

        ``g_names`` (:obj:`dict`):  A dictionary of graphviz names.\
         This dictionary is bound to the function by the calling routine.

    Returns:
        A :obj:`str`.  The node label.

    """
    return g_names[name]


def make_zone_node_label(name, zone, zone_label, g_names):
    """The default zone node label function.

    Args:
        ``name`` (:obj:`str`):  The species name.

        ``zone``:  A `wnutils <https://wnutils.readthedocs.io>`_ zone object.

        ``zone_label`` (:obj:`str` or :obj:`tuple`):  The label of the zone.

        ``g_names`` (:obj:`dict`):  A dictionary of graphviz names.
        This dictionary is bound to the function by the calling routine.

    Returns:
        A :obj:`str`.  The node label.

    """
    assert zone
    assert zone_label

    return make_node_label(name, g_names)


def scale_edge_weight(edge_data, f_max, scale, threshold):
    """The default edge weight scale function.

    Args:
        ``edge_data``:  A dictionary of edge properties.

        ``f_max`` (:obj:`float`):  The maximum edge weight.

        ``scale`` (:obj:`float`):  The factor by which to scale the weight.

        ``threshold`` (:obj:`float`):  The threshold for not including the edge.

    Returns:
        A :obj:`bool` indicating whether to include the edge (True) or not (False).

    """

    if f_max == 0:
        return False

    if "penwidth" in edge_data:
        return True

    keep_edge = True
    _r = edge_data["weight"] / f_max
    if _r >= threshold:
        edge_data["penwidth"] = scale * _r
    else:
        keep_edge = False
    return keep_edge


def create_flow_graph(net, t_9, rho, mass_fractions, **kwargs):
    """A routine to create a flow graph for a given set of mass fractions at
       the input temperature and density.

    Args:
        ``net``: A wnnet network. 

        ``t_9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which to\
         compute the flows. 

        ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
          flows.
        
        ``mass_fractions`` (:obj:`float`): A\
          `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
          fractions.

        ``**kwargs``: The allowed optional\
          :ref:`keyword <Allowed_keywords>` arguments for this routine are:\
          `induced_nuc_xpath`_, `induced_reac_xpath`_, `flow_type`_,\
          `user_funcs`_, `reaction_color_tuples`_, `threshold`_, `scale`_,\
          `state_scaling`_, `allow_isolated_species`_, `title_func`_,\
          `node_label_func`_, `scale_edge_weight_func`_, `graph_attributes`_,\
          `edge_attributes`_, `node_attributes`_, `solar_species`_,\
          `solar_node_attributes`_, `special_node_attributes`_,\
          `special_edge_attributes`_.


    Returns:
        A `networkx multidigraph \
        <https://networkx.org/documentation/stable/reference/classes/multidigraph.html>`_\
        showing the flows.
     

    """

    my_list = [
        "induced_nuc_xpath",
        "induced_reac_xpath",
        "flow_type",
        "user_funcs",
        "reaction_color_tuples",
        "threshold",
        "scale",
        "state_scaling",
        "allow_isolated_species",
        "title_func",
        "node_label_func",
        "scale_edge_weight_func",
        "graph_attributes",
        "edge_attributes",
        "node_attributes",
        "solar_species",
        "solar_node_attributes",
        "special_node_attributes",
        "special_edge_attributes",
    ]

    my_args = gh.get_keywords(my_list, **kwargs)

    assert my_args["flow_type"] in ("net", "full")

    my_flows = wf.compute_flows(
        net,
        t_9,
        rho,
        mass_fractions,
        reac_xpath=my_args["induced_reac_xpath"],
        user_funcs=my_args["user_funcs"],
    )

    # Get the subset of nuclides to view in the graph.  Get anchors.

    subset_nuclides, anchors = gh.get_subset_and_anchors(
        net.get_nuclides(nuc_xpath=my_args["induced_nuc_xpath"])
    )

    # Title

    if not my_args["title_func"]:
        my_args["title_func"] = functools.partial(
            make_t9_rho_flow_string, t_9=t_9, rho=rho
        )

    # Node label

    if not my_args["node_label_func"]:
        g_names = net.xml.get_graphviz_names(subset_nuclides)
        my_args["node_label_func"] = functools.partial(
            make_node_label, g_names=g_names
        )

    # Scale edge weight

    if not my_args["scale_edge_weight_func"]:
        my_args["scale_edge_weight_func"] = scale_edge_weight

    return gh.create_flow_graph(
        net, my_flows, subset_nuclides, anchors, **my_args
    )


def create_zone_flow_graphs(net, zones, **kwargs):
    """A routine to create flow graphs for a set of zones.

    Args:
        ``net``: A wnnet network. 

        ``zones``: A `wnutils <https://wnutils.readthedocs.io>`_\
         dictionary of zones.

        ``**kwargs``: The allowed optional\
          :ref:`keyword <Allowed_keywords>` arguments for this routine are:\
          `induced_nuc_xpath`_, `induced_reac_xpath`_, `flow_type`_,\
          `zone_user_funcs`_, `reaction_color_tuples`_, `threshold`_,\
          `scale`_, `state_scaling`_, `allow_isolated_species`_,\
          `zone_title_func`_, `zone_node_label_func`_,\
          `scale_edge_weight_func`_,\
          `graph_attributes`_, `edge_attributes`_, `node_attributes`_,\
          `solar_species`_, `solar_node_attributes`_,\
          `special_node_attributes`_, `special_edge_attributes`_.

    Returns:
        A :obj:`dict` of\
        `networkx multidigraphs\
         <https://networkx.org/documentation/stable/reference/classes/multidigraph.html>`_\
         showing the flows.  The keys are the zone labels.
     
    """

    my_list = [
        "induced_nuc_xpath",
        "induced_reac_xpath",
        "flow_type",
        "zone_user_funcs",
        "reaction_color_tuples",
        "threshold",
        "scale",
        "state_scaling",
        "allow_isolated_species",
        "zone_title_func",
        "zone_node_label_func",
        "scale_edge_weight_func",
        "graph_attributes",
        "edge_attributes",
        "node_attributes",
        "solar_species",
        "solar_node_attributes",
        "special_node_attributes",
        "special_edge_attributes",
    ]

    my_args = gh.get_keywords(my_list, **kwargs)

    result = {}

    my_flows = wf.compute_flows_for_zones(
        net,
        zones,
        reac_xpath=my_args["induced_reac_xpath"],
        user_funcs=my_args["zone_user_funcs"],
    )

    subset_nuclides, anchors = gh.get_subset_and_anchors(
        net.get_nuclides(nuc_xpath=my_args["induced_nuc_xpath"])
    )

    g_names = net.xml.get_graphviz_names(subset_nuclides)

    if not my_args["scale_edge_weight_func"]:
        my_args["scale_edge_weight_func"] = scale_edge_weight

    # Loop on zones

    for key, value in zones.items():

        # Title

        if not my_args["zone_title_func"]:
            my_args["title_func"] = functools.partial(
                make_time_t9_rho_flow_string, zone=value, zone_label=key
            )
        else:
            my_args["title_func"] = functools.partial(
                my_args["zone_title_func"], zone=value, zone_label=key
            )

        # Node label

        if not my_args["zone_node_label_func"]:
            my_args["node_label_func"] = functools.partial(
                make_zone_node_label,
                zone=value,
                zone_label=key,
                g_names=g_names,
            )
        else:
            my_args["node_label_func"] = functools.partial(
                my_args["zone_node_label_func"], zone=value, zone_label=key
            )

        # Create graph

        result[key] = gh.create_flow_graph(
            net, my_flows[key], subset_nuclides, anchors, **my_args
        )

    return result


def create_network_graph(net, **kwargs):
    """A routine to create a network graph showing species and reactions\
        among them.

    Args:
        ``net``: A wnnet network. 

        ``**kwargs``: The allowed optional\
          :ref:`keyword <Allowed_keywords>` arguments for this routine are:\
          `induced_nuc_xpath`_, `induced_reac_xpath`_, `direction`_,\
          `reaction_color_tuples`_, `threshold`_, `scale`_, `state_scaling`_,\
          `allow_isolated_species`_, `node_label_func`_, `graph_attributes`_,\
          `edge_attributes`_, `node_attributes`_, `solar_species`_,\
          `solar_node_attributes`_, `special_node_attributes`_,\
          `special_edge_attributes`_.

    Returns:
        A `networkx multidigraph\
        <https://networkx.org/documentation/stable/reference/classes/multidigraph.html>`_\
        showing the network; that is, the nuclides and reactions among them.

    """

    my_list = [
        "induced_nuc_xpath",
        "induced_reac_xpath",
        "direction",
        "reaction_color_tuples",
        "threshold",
        "scale",
        "state_scaling",
        "allow_isolated_species",
        "node_label_func",
        "graph_attributes",
        "node_attributes",
        "edge_attributes",
        "solar_species",
        "solar_node_attributes",
        "special_node_attributes",
        "special_edge_attributes",
    ]

    my_args = gh.get_keywords(my_list, **kwargs)

    assert my_args["direction"] in ("forward", "reverse", "both")

    nuclides = net.get_nuclides()

    # Get the subset of nuclides to view in the graph.  Get anchors.

    val, anchors = gh.get_subset_and_anchors(
        net.get_nuclides(nuc_xpath=my_args["induced_nuc_xpath"])
    )

    d_g = nx.MultiDiGraph()

    # Add nodes.

    for nuc in nuclides:
        d_g.add_node(nuc)

    # Add arcs (reactions).

    gh.add_reactions_to_graph(net, d_g, my_args)

    # Apply attributes

    gh.apply_graph_attributes(d_g, my_args["graph_attributes"])

    gh.apply_node_attributes(d_g, my_args["node_attributes"])

    gh.apply_edge_attributes(d_g, my_args["edge_attributes"])

    gh.apply_solar_node_attributes(
        d_g, my_args["solar_species"], my_args["solar_node_attributes"]
    )

    gh.apply_special_node_attributes(d_g, my_args["special_node_attributes"])

    gh.apply_special_edge_attributes(d_g, my_args["special_edge_attributes"])

    # Remove isolated nodes if desired

    if not my_args["allow_isolated_species"]:
        gh.remove_isolated_nodes(d_g, my_args)

    # Restore anchors

    for anchor in anchors:
        if anchor not in d_g.nodes:
            d_g.add_node(anchor, style="invis")

    # Subgraph

    sub_graph = nx.subgraph(d_g, val)

    # Node label

    if my_args["node_label_func"]:
        _node_label_func = my_args["node_label_func"]
    else:
        g_names = net.xml.get_graphviz_names(list(sub_graph.nodes.keys()))
        _node_label_func = functools.partial(make_node_label, g_names=g_names)

    for node in sub_graph.nodes:
        sub_graph.nodes[node]["pos"] = gh.get_pos(
            net, node, my_args["state_scaling"]
        )
        sub_graph.nodes[node]["label"] = _node_label_func(node)

    gh.color_edges(sub_graph, net, my_args["reaction_color_tuples"])

    return sub_graph


def create_zone_integrated_current_graphs(net, zones, **kwargs):
    """A routine to create integrated currents graphs for a set of zones.

    Args:
        ``net``: A wnnet network. 

        ``zones``: A `wnutils <https://wnutils.readthedocs.io>`_ dictionary\
          of zones.

        ``**kwargs``: The allowed optional\
          :ref:`keyword <Allowed_keywords>` arguments for this routine are:\
          `induced_nuc_xpath`_, `induced_reac_xpath`_,\
          `reaction_color_tuples`_, `threshold`_, `scale`_, `state_scaling`_,\
          `allow_isolated_species`_, `zone_title_func`_,\
          `zone_node_label_func`_,\
          `scale_edge_weight_func`_, `graph_attributes`_, `edge_attributes`_,\
          `node_attributes`_, `solar_species`_, `solar_node_attributes`_,\
          `special_node_attributes`_, `special_edge_attributes`_.

    Returns:
        A :obj:`dict` of `networkx multidigraphs \
        <https://networkx.org/documentation/stable/reference/classes/multidigraph.html>`_ \
        showing the integrated currents.  The keys are the zone labels.

    """

    my_list = [
        "induced_nuc_xpath",
        "induced_reac_xpath",
        "reaction_color_tuples",
        "threshold",
        "scale",
        "state_scaling",
        "allow_isolated_species",
        "zone_title_func",
        "zone_node_label_func",
        "scale_edge_weight_func",
        "graph_attributes",
        "edge_attributes",
        "node_attributes",
        "solar_species",
        "solar_node_attributes",
        "special_node_attributes",
        "special_edge_attributes",
    ]

    my_args = gh.get_keywords(my_list, **kwargs)

    result = {}

    subset_nuclides, anchors = gh.get_subset_and_anchors(
        net.get_nuclides(nuc_xpath=my_args["induced_nuc_xpath"])
    )

    g_names = net.xml.get_graphviz_names(list(net.get_nuclides().keys()))

    if not my_args["scale_edge_weight_func"]:
        my_args["scale_edge_weight_func"] = scale_edge_weight

    for key, value in zones.items():

        # Title

        if not my_args["zone_title_func"]:
            my_args["title_func"] = functools.partial(
                make_time_t9_rho_current_string, zone=value, zone_label=key
            )
        else:
            my_args["title_func"] = functools.partial(
                my_args["zone_title_func"], zone=value, zone_label=key
            )

        # Node label

        if not my_args["zone_node_label_func"]:
            my_args["node_label_func"] = functools.partial(
                make_zone_node_label,
                zone=value,
                zone_label=key,
                g_names=g_names,
            )
        else:
            my_args["node_label_func"] = functools.partial(
                my_args["zone_node_label_func"],
                zone=value,
                zone_label=key,
                g_names=g_names,
            )

        result[key] = gh.create_integrated_current_graph(
            net, value, subset_nuclides, anchors, **my_args
        )

    return result



"""This module contains constants for use in the wnnet package.  The module
also includes all constants from
`gslconsts <https://gslconsts.readthedocs.io/en/latest/gslconsts.html#module-gslconsts.consts>`_.
The gsl constants are imported into their corresponding gslconsts namespaces;
thus, for example, the speed of light in vacuum in MKSA units would be
*wnnet.consts.consts.GSL_CONST_MKSA_SPEED_OF_LIGHT* while
the Euler-Mascheroni constant would be *wnnet.consts.math.M_EULER*.
"""

import numpy as np
from gslconsts import consts

# Conversions for use with the package.

#: Speed of light in vacuum (cm/s)
c = consts.GSL_CONST_CGSM_SPEED_OF_LIGHT

#: Boltzmann's constant (ergs/K)
k_B = consts.GSL_CONST_CGSM_BOLTZMANN

#: Electron mass (g)
m_e = consts.GSL_CONST_CGSM_MASS_ELECTRON

#: Amu mass (g)
u = consts.GSL_CONST_CGSM_UNIFIED_ATOMIC_MASS

#: Avogadro's number
N_A = consts.GSL_CONST_NUM_AVOGADRO

#: Planck's constant divided by 2pi (ergs/s)
hbar = consts.GSL_CONST_CGSM_PLANCKS_CONSTANT_HBAR

#: Conversion factor from MeVs to ergs
MeV_to_ergs = 1.0e6 * consts.GSL_CONST_CGSM_ELECTRON_VOLT

#: Conversion factor from ergs to MeVs
ergs_to_MeV = 1.0 / MeV_to_ergs

#: Electron rest mass (MeV)
m_e_in_MeV = m_e * np.power(c, 2) * ergs_to_MeV

#: Amu rest mass (MeV)
m_u_in_MeV = u * np.power(c, 2) * ergs_to_MeV



"""This module handles `webnucleo <https://webnucleo.readthedocs.io>`_ nuclear reaction networks."""

import numpy as np
import wnnet.nuc as wn
import wnnet.reac as wr
import wnnet.consts as wc


class Net(wn.Nuc, wr.Reac):
    """A class to store webnucleo networks.

    Args:
        ``file`` (:obj:`str`): A string giving the name of the XML file with\
          the network data.

        ``nuc_xpath`` (:obj:`str`, optional): An XPath expression to select\
         nuclides.  Default is all nuclides.

        ``reac_xpath`` (:obj:`str`, optional):  An XPath expression to select\
        reactions.  Default is all reactions.

    """

    def __init__(self, file, nuc_xpath="", reac_xpath=""):
        wn.Nuc.__init__(self, file, nuc_xpath=nuc_xpath)
        wr.Reac.__init__(self, file, reac_xpath=reac_xpath)
        self.valid_reactions = {}
        self.valid_reactions[("", "")] = self.get_valid_reactions(
            nuc_xpath=nuc_xpath, reac_xpath=reac_xpath
        )

    def compute_q_values(self, nuc_xpath="", reac_xpath=""):
        """A method to compute reaction Q values for valid reactions in the network.

        Args:
            ``nuc_xpath`` (:obj:`str`, optional): An XPath expression to select\
            the nuclides.

            ``reac_xpath`` (:obj:`str`, optional): An XPath expression to\
            select the reactions.

        Returns:
            A :obj:`dict`.  The keys for the dictionary are the reaction\
            strings and the values are the corresponding Q values.

        """

        result = {}

        for reac in self.get_valid_reactions(
            nuc_xpath=nuc_xpath, reac_xpath=reac_xpath
        ):
            tmp = self.compute_reaction_q_value(reac)
            if tmp:
                result[reac] = tmp

        return result

    def get_valid_reactions(self, nuc_xpath="", reac_xpath=""):
        """Method to retrieve the valid reactions in the network.

        Args:
            ``nuc_xpath`` (:obj:`str`, optional):  An XPath expression to\
              select nuclides.  Default is all nuclides.

            ``reac_xpath`` (:obj:`str`, optional):  An XPath expression to\
              select reactions.  Default is all reactions.

        Returns:
            A :obj:`dict` of `wnutils <https://wnutils.readthedocs.io>`_ \
              reactions.

        """

        if (nuc_xpath, reac_xpath) not in self.valid_reactions:
            result = {}
            reactions = self.get_reactions(reac_xpath=reac_xpath)
            for reac in reactions:
                if self.is_valid_reaction(reac, nuc_xpath=nuc_xpath):
                    result[reac] = reactions[reac]
            self.valid_reactions[(nuc_xpath, reac_xpath)] = result

        return self.valid_reactions[(nuc_xpath, reac_xpath)]

    def compute_reaction_q_value(self, name):
        """Method to compute the Q value for a reaction.

        Args:
            ``name`` (:obj:`str`):  A string giving the reaction.

        Returns:
            A :obj:`float` giving the Q value for the reaction or None if the\
             reaction is not valid.

        """

        nuclides = self.get_nuclides()
        reaction = self.get_reactions()[name]
        result = 0
        for spec in reaction.nuclide_reactants:
            if spec not in nuclides:
                return None
            result += nuclides[spec]["mass excess"]
        for spec in reaction.nuclide_products:
            if spec not in nuclides:
                return None
            result -= nuclides[spec]["mass excess"]

        if (
            "positron" in reaction.products
            and "neutrino_e" in reaction.products
        ):
            result -= 2.0 * wc.m_e_in_MeV

        return result

    def _obeys_conservation_laws(self, nuclides, reaction):
        e_p = {"electron": -1, "positron": 1}
        e_lepton = {
            "electron": 1,
            "positron": -1,
            "neutrino_e": 1,
            "anti-neutrino_e": -1,
        }
        mu_lepton = {"neutrino_mu": 1, "anti-neutrino_mu": -1}
        tau_lepton = {"neutrino_tau": 1, "anti-neutrino_tau": -1}

        d_z = 0
        d_a = 0
        d_l_e = 0
        d_l_mu = 0
        d_l_tau = 0

        for spec in reaction.reactants:
            if spec in nuclides:
                d_a += nuclides[spec]["a"]
                d_z += nuclides[spec]["z"]
            if spec in e_p:
                d_z += e_p[spec]
            if spec in e_lepton:
                d_l_e += e_lepton[spec]
            if spec in mu_lepton:
                d_l_mu += mu_lepton[spec]
            if spec in tau_lepton:
                d_l_tau += tau_lepton[spec]
        for spec in reaction.products:
            if spec in nuclides:
                d_a -= nuclides[spec]["a"]
                d_z -= nuclides[spec]["z"]
            if spec in e_p:
                d_z -= e_p[spec]
            if spec in e_lepton:
                d_l_e -= e_lepton[spec]
            if spec in mu_lepton:
                d_l_mu -= mu_lepton[spec]
            if spec in tau_lepton:
                d_l_tau -= tau_lepton[spec]

        return (
            d_a == 0
            and d_z == 0
            and d_l_e == 0
            and d_l_mu == 0
            and d_l_tau == 0
        )

    def is_valid_reaction(self, name, nuc_xpath=""):
        """Method to determine a reaction is valid in the network.

        Args:
            ``name`` (:obj:`str`):  A string giving the reaction.

            ``nuclides`` (:obj:`str`, optional):  An XPath expression\
              to select nuclides.  Default is all nuclides.

        Returns:
            A :obj:`bool` with value True if the reaction is valid and False\
            if not.

        """

        nuclides = self.get_nuclides(nuc_xpath=nuc_xpath)
        reaction = self.get_reactions()[name]

        if not self._obeys_conservation_laws(nuclides, reaction):
            return False

        for spec in reaction.nuclide_reactants + reaction.nuclide_products:
            if spec not in nuclides:
                return False
        return True

    def compute_rates_for_reaction(self, name, t_9, user_funcs=""):
        """Method to compute the forward and reverse rates for a valid reaction.

        Args:
            ``name`` (:obj:`str`):  A string giving the reaction.

            ``t_9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at\
             which to compute the rates.

            ``user_funcs`` (:obj:`dict`, optional): A dictionary of\
             user-defined functions associated with a user_rate key.

        Returns:
            A two-element :obj:`tuple` with the first element being the\
            forward rate and the second element being the reverse rate.\
            If the reaction is not valid, returns None.

        """

        if not self.is_valid_reaction(name):
            return None

        reaction = self.get_reactions()[name]
        forward = reaction.compute_rate(t_9, user_funcs=user_funcs)

        if self.is_weak_reaction(name) or len(reaction.reactants) == 1:
            return (forward, 0)

        d_exp = 0

        for spec in reaction.nuclide_reactants:
            d_exp += self.compute_nse_factor(spec, t_9, 1.0)
        for spec in reaction.nuclide_products:
            d_exp -= self.compute_nse_factor(spec, t_9, 1.0)

        if d_exp < -300.0:
            return (forward, 0)

        if d_exp > 300.0:
            return (0, 0)

        tup = self.compute_reaction_duplicate_factors(name)

        return (forward, np.exp(d_exp) * (tup[1] / tup[0]) * forward)

    def compute_rates(self, t_9, nuc_xpath="", reac_xpath="", user_funcs=""):
        """Method to compute the forward and reverse rates for valid\
           reactions in a network.

        Args:
            ``t_9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at\
              which to compute the rates.

            ``nuc_xpath`` (:obj:`str`, optional):  An XPath expression\
              to select nuclides.  Default is all nuclides.

            ``reac_xpath`` (:obj:`str`, optional):  An XPath expression\
              to select reactions.  Default is all reactions.

            ``user_funcs`` (:obj:`dict`, optional): A dictionary of\
              user-defined functions associated with a user_rate key.

        Returns:
            A :obj:`dict` containing the rates.  The key is the reaction\
            string while the value is a two-element  :obj:`tuple` with the\
            first element being the forward rate and the second element\
            being the reverse rate.

        """

        v_reactions = self.get_valid_reactions(
            nuc_xpath=nuc_xpath, reac_xpath=reac_xpath
        )

        result = {}

        for reac in v_reactions:
            result[reac] = self.compute_rates_for_reaction(
                reac, t_9, user_funcs=user_funcs
            )

        return result



"""
A package of python routines to handle webnucleo networks.
"""

__all__ = [
    "__title__",
    "__summary__",
    "__version__",
    "__author__",
    "__copyright__",
]

__title__ = "wnnet"
__summary__ = "Python project to handle webnucleo reaction networks"
__version__ = "2.6.1"
__author__ = "Clemson University"
__copyright__ = "Clemson University, 2022-2025"
